import re
from inspect import signature
from typing import Callable

from m.core import Res
from m.pydantic import KebabModel
from pydantic import BaseModel
from pydantic.fields import FieldInfo

InputOutputs = tuple[type[KebabModel], type[KebabModel]]


class InputIssue(BaseModel):
    """An issue with an input."""

    # Set if the value is invalid, may be set of the form "input_name: value"
    invalid_input_value: str


class MetadataOutput(BaseModel):
    """An action metadata file output."""

    description: str
    value: str  # noqa: WPS110 - Actual name of property, cannot change.


def get_inputs_outputs(
    run_function: Callable[[KebabModel], Res[KebabModel]],
) -> InputOutputs:
    """Get the inputs and outputs of a `run_function`.

    Args:
        run_function: The function to obtain the inputs and outputs from.

    Returns:
        A tuple of the inputs and outputs.
    """
    func_sig = signature(run_function)
    in_params = func_sig.parameters
    arg_name = next(iter(in_params))
    input_model = in_params[arg_name].annotation
    out_sig = func_sig.return_annotation
    # expecting out_sig to be of type Res[Model] which is OneOf
    output_model: type[KebabModel] = out_sig.__args__[0].__args__[1]
    return input_model, output_model


def map_args(
    all_args: dict[str, str],
    available_values: dict[str, str],
    dict_key: Callable[[str], str],
) -> dict[str, str]:
    """Map the arguments to a new value using the available values.

    Args:
        all_args: A dictionary with mappable values.
        available_values: A dictionary with available values.
        dict_key: A transform function for the new key in the map.

    Returns:
        A mapped dictionary.
    """
    return {
        dict_key(key): (
            expression_eval(available_values[arg_val])
            if arg_val in available_values
            else arg_val
        )
        for key, arg_val in all_args.items()
    }


def is_input_key(key: str) -> bool:
    """Determine if a key is of the form `(step_id).(arg_name)`.

    Args:
        key: The key to check.

    Returns:
        True if the key is an input key.
    """
    pattern = r'^[\w-]+\.[\w-]+$'
    return bool(re.match(pattern, key))


def is_export(field: FieldInfo) -> bool:
    """Determine if a field is exported.

    Args:
        field: The field to check.

    Returns:
        Whether the field is exported.
    """
    if isinstance(field.json_schema_extra, dict):
        return bool(field.json_schema_extra.get('export', False))
    # internally we do not set json_schema_extra to something other than dict
    # If we are reaching the next step it is because of a custom field.
    return False  # pragma: no cover


def input_env(name: str) -> str:
    """Get the name of the env variable for an input.

    Args:
        name: The name of the input.

    Returns:
        The name of the env variable.
    """
    upper_name = name.replace('-', '_').upper()
    return f'INPUT_{upper_name}'


def expression_eval(gh_expression: str) -> str:
    """Create a string to evaluate a Github expression.

    Args:
        gh_expression: The Github expression to evaluate.

    Returns:
        A string to evaluate the Github expression.
    """
    left = '{{'
    right = '}}'
    return f'${left} {gh_expression} {right}'


def verify_inputs(
    input_model: type[KebabModel],
    args: dict[str, str],
    available_outputs: dict[str, str],
) -> list[InputIssue]:
    """Verify the inputs.

    This is used to make sure we declare valid arguments in each step
    along with valid accessible values from the global inputs or previous
    steps.

    Args:
        input_model: The model to use to verify the inputs.
        args: The arguments to verify.
        available_outputs: A dict of available outputs to use as inputs.

    Returns:
        A list of issues.
    """
    issues: list[InputIssue] = []
    for prop_name in input_model.model_fields:
        input_name = prop_name
        arg_value = args.get(input_name)
        not_available = arg_value not in available_outputs
        if arg_value and not_available and is_input_key(arg_value):
            issues.append(
                InputIssue(
                    invalid_input_value=f'{input_name}={arg_value}',
                ),
            )
    return issues


class OutputField(BaseModel):
    """Digestible information about an output."""

    # name of the output
    name: str

    # description of the output
    description: str

    # steps.{step_id}.outputs.{output_name}
    full_ref_name: str

    # {step_id}.{output_name}
    short_ref_name: str

    # flag to indicate if this output should be exported
    is_export: bool

    def get_metadata_output(self: 'OutputField') -> MetadataOutput:
        """Get the metadata output.

        Returns:
            An instance of the metadata output.
        """
        return MetadataOutput(
            description=self.description,
            value=expression_eval(self.full_ref_name),
        )

    @classmethod
    def create(
        cls: type['OutputField'],
        step_id: str,
        output_model: type[KebabModel],
    ) -> dict[str, 'OutputField']:
        """Create an instance of an output field.

        Args:
            step_id: The id of the step.
            output_model: The model to create it from.

        Returns:
            An instance of the output field.
        """
        outputs: dict[str, OutputField] = {}
        for name, prop_info in output_model.model_fields.items():
            kebab_name = prop_info.alias or name
            outputs[name] = OutputField(
                name=name,
                description=prop_info.description or '',
                full_ref_name=f'steps.{step_id}.outputs.{kebab_name}',
                short_ref_name=f'{step_id}.{name}',
                is_export=is_export(prop_info),
            )
        return outputs


# first item is the action outputs and second is the collection of all
# outputs that can be used when defining inputs in steps.
ActionOutputs = tuple[dict[str, MetadataOutput], dict[str, str]]



from m.cli import Arg, BaseModel, command, run_main
from m.core.io import env


class Arguments(BaseModel):
    r"""Provide the build sha.

    Given the GITHUB_SHA (a merge commit it fetches the sha of the actual
    commit we thought we were building.

    example::

        $ m github build_sha \
            --owner jmlopez-rod \
            --repo m \
            --sha 6bf3a8095891c551043877b922050d5b01d20284
        fa6a600729ffbe1dfd7fece76ef4566e45fbfe40

    The sha can be obtained in Github by looking at the output of the
    checkout action.

    The `owner` option defaults to the value of the environment variable
    `GITHUB_REPOSITORY_OWNER`.
    """

    owner: str = Arg(
        default=env('GITHUB_REPOSITORY_OWNER'),
        help='repo owner',
    )
    repo: str = Arg(
        help='repo name',
        required=True,
    )
    sha: str = Arg(
        help='commit sha',
        required=True,
    )


@command(
    help='get the correct commit sha',
    model=Arguments,
)
def run(arg: Arguments, arg_ns) -> int:
    from m.github.ci import get_build_sha
    return run_main(
        lambda: get_build_sha(arg_ns.token, arg.owner, arg.repo, arg.sha),
        result_handler=print,
    )



import json
from collections.abc import Mapping
from contextlib import suppress
from typing import Any
from typing import Mapping as Map
from typing import Union

from . import rw
from .fp import Good, OneOf
from .issue import Issue
from .one_of import hone, issue, one_of


def read_json(
    filename: str | None,
    error_if_empty: bool = False,
) -> OneOf[Issue, Any]:
    """Read a json object from a json file.

    Args:
        filename: The filename to read from, if `None` it reads from stdin.
        error_if_empty: The json parser may throw an error if `True`.

    Returns:
        A `Good` containing the parsed contents of the json file.
    """
    empty: str = '' if error_if_empty else 'null'
    context = {'filename': filename or 'SYS.STDIN'}
    return one_of(lambda: [
        json_data
        for json_str in rw.read_file(filename)
        for json_data in parse_json(json_str or empty, error_if_empty)
    ]).flat_map_bad(hone('failed to read json file', context=context))


def parse_json(
    json_str: str,
    error_if_empty: bool = False,
) -> OneOf[Issue, Any]:
    """Parse a string as json.

    Args:
        json_str: The string to parse.
        error_if_empty: The json parser may throw an error if `True`.

    Returns:
        A `Good` containing the parsed contents of the json string.
    """
    empty = '' if error_if_empty else 'null'
    try:
        return Good(json.loads(json_str or empty))
    except Exception as ex:
        return issue('failed to parse the json data', cause=ex)


def get(dict_inst: Any, key_str: str) -> OneOf[Issue, Any]:
    """Return the value based on the `key_str` specified.

    The following are equivalent::

        obj['a']['b']['c'] <=> get(obj, 'a.b.c').value

    provided that the keys `a`, `b` and `c` are valid.

    Args:
        dict_inst: The dictionary instance to query.
        key_str: A simple query to fetch from the dict_inst.

    Returns:
        A `Good` if the value we want is available, otherwise it returns a
        `Bad` with the path that returned `None`.
    """
    keys = key_str.split('.')
    current = dict_inst
    for num, key in enumerate(keys):
        new_key: Union[str, int] = key
        with suppress(ValueError):
            new_key = int(key)
        try:
            current = current[new_key]
        except KeyError:
            pth = '.'.join(keys[:num + 1])
            return issue(f'`{pth}` path was not found')
        except Exception as ex:
            pth = '.'.join(keys[:num])
            if not isinstance(current, Mapping):
                context = pth or current
                return issue(f'`{context}` is not a dict')
            # catch unknown issue
            return issue(  # pragma: no cover
                f'{pth} resulted in an error',
                cause=ex,
            )
    return Good(current)


def multi_get(
    dict_inst: object,
    *keys: str,
) -> OneOf[Issue, list[Any]]:
    """Call `get` for every input specified by `keys`.

    It collects the invalid keys and returns an `Issue`::

        multi_get(obj, 'a', 'a.b', 'a.b.c')

    Args:
        dict_inst: The dictionary instance to query.
        keys: The queries to apply.

    Returns:
        A `Good` with a list of the results if successful, otherwise a `Bad`
        with the list of failures.
    """
    result_items = []
    failures = []
    for key in keys:
        res = get(dict_inst, key)
        if res.is_bad:
            failures.append(
                Issue(
                    message=f'key lookup failure: `{key}`',
                    cause=res.value,
                    include_traceback=False,
                ),
            )
        else:
            result_items.append(res.value)
    if failures:
        return issue(
            'multi_get key retrieval failure',
            context=[x.to_dict() for x in failures],
            include_traceback=False,
        )
    return Good(result_items)


def _to_str(inst: Any):
    if isinstance(inst, bool):
        return f'{inst}'.lower()
    if inst is None:
        return 'null'
    return f'{inst}'


def jsonq(
    dict_inst: Map[str, Any],
    separator: str,
    *key_str: str,
) -> OneOf[Issue, str]:
    """Stringify the values obtained from `multi_get`.

    Args:
        dict_inst: The dictionary instance to query.
        separator: A string to use to separate the results.
        key_str: The queries to apply.

    Returns:
        A string separated by `separator` if successful or an Issue with
        a description of the problems.
    """
    return one_of(lambda: [
        separator.join(map(_to_str, key_values))
        for key_values in multi_get(dict_inst, *key_str)
    ])



from m.cli import Arg, ArgProxy, BaseModel, command, run_main


class Arguments(BaseModel):
    """Prompt user for the next valid semantic version."""

    type: str = ArgProxy(
        '--type',
        required=True,
        choices=['release', 'hotfix'],
        help='verification type',
    )
    version: str = Arg(
        help='version to bump',
        positional=True,
        required=True,
    )


@command(
    help='prompt for the next version',
    model=Arguments,
)
def run(arg: Arguments):
    from m.core.io import prompt_next_version

    return run_main(
        lambda: prompt_next_version(arg.version, arg.type),
        result_handler=print,
    )



import logging

from m.core import io as mio

from .formatters import CiFormatter, JsonFormatter
from .handlers import JsonFileHandler, StdErrHandler, StdOutHandler


def logging_config(level: int | None = None, json_file: str = '') -> None:
    """Apply a configuration to the logs.

    Args:
        level: The logging level, defaults to INFO.
        json_file: Optional file name where to store each log record as json.
    """
    formatter = CiFormatter()
    stdout_handler = StdOutHandler(formatter)
    stderr_handler = StdErrHandler(formatter)
    all_handlers = [stderr_handler, stdout_handler]
    if json_file:
        json_formatter = JsonFormatter()
        all_handlers.append(JsonFileHandler(json_formatter, json_file))

    debug_logs = mio.env('DEBUG_M_LOGS', 'false') == 'true'
    default_level = logging.DEBUG if debug_logs else logging.INFO
    logging_level = level if level is not None else default_level
    logging.basicConfig(
        level=logging_level,
        handlers=all_handlers,
        force=True,
    )



from m.cli import Arg, BaseModel, command, run_main, validate_file_exists


class Arguments(BaseModel):
    """Create the actions metadata as described by given python script.

    When creating actions we need to have the `python_file` be at the root
    of a python project. This is because when we run the action we will set
    the path to the directory containing the file.

    By default it has been set to `src/actions.py` since this is a common case
    for python projects.

    The `actions.py` can be name anything but it must contain a variable named
    `actions`. This variable must be an instance or list of
    [m.github.actions.Action][].
    """

    python_file: str = Arg(
        default='src/actions.py',
        help='Python script containing the `actions` variable.',
        positional=True,
        validator=validate_file_exists,
    )

    check: bool = Arg(
        default=False,
        help='Check if generated files are up to date and skip writing them.',
    )

    show_traceback: bool = Arg(
        default=False,
        help='Display traceback information in case of an error.',
    )


@command(
    help='create actions metadata',
    model=Arguments,
)
def run(arg: Arguments) -> int:
    from m.github.actions.builder import build_actions

    return run_main(
        lambda: build_actions(
            actions_py_file=arg.python_file,
            check=arg.check,
            show_traceback=arg.show_traceback,
        ),
        result_handler=lambda _: None,
    )



import argparse as ap
from dataclasses import dataclass
from typing import Any, Callable

import typing_extensions
from pydantic import BaseModel
from pydantic_core import PydanticUndefined


@dataclass
class CommandInputs:
    """Inputs to command decorator."""

    help: str

    model: type[BaseModel]


@dataclass
class FuncArgs:
    """Stores function arguments."""

    args: list[Any]
    kwargs: dict[str, Any]


@typing_extensions.deprecated(
    'The `add_arg` method is deprecated; use `m.cli.ArgProxy` instead.',
)
def add_arg(*args: Any, **kwargs: Any) -> FuncArgs:
    """Wrap FuncArgs arguments in a function.

    Args:
        args: The arguments to argparse add_arguments.
        kwargs: The keyword arguments to argparse add arguments.

    Returns:
        A FuncArgs instance.
    """
    # `m` does not reference this function anymore, excluding from coverage
    return FuncArgs(args=list(args), kwargs=kwargs)  # pragma: no cover


# This is the actual signature of the run function after the `command`
# annotation is set on it.
DecoratedRunFunction = Callable[
    [str, ap.Namespace | None, ap._SubParsersAction | None],  # noqa: WPS437, WPS465 # pylint:disable=protected-access
    int,
]


@dataclass
class CommandModule:
    """Container to store the run function from a "command" module."""

    run: DecoratedRunFunction


CommandModuleMap = dict[str, CommandModule]


@dataclass
class MetaModule:
    """Container to store a metadata dictionary from a "meta" module."""

    meta: dict[str, str]
    add_arguments: Callable[[ap.ArgumentParser], None] | None = None


@dataclass
class CliSubcommands:
    """Container to store subcommands."""

    # Dictionary of subcommands.
    subcommands: CommandModuleMap

    # Each subcommand needs to provide metadata to create the help message.
    meta: MetaModule | None


@dataclass
class CliCommands:
    """Container to store the commands and subcommands for the cli."""

    commands: dict[str, CommandModule | CliSubcommands]

    # Optional root meta data to provide information about the cli.
    meta: MetaModule | None


MISSING = PydanticUndefined
AnyMap = dict[str, Any]



from m.cli import Arg, ArgProxy, BaseModel, command, run_main


class Arguments(BaseModel):
    """Fail command when working on a non valid branch for a release/hotfix.

    Used during a release setup or hotfix setup. We want to make sure
    that we are working on the correct branch depending on release type
    we want to make and the workflow that we are using.
    """

    type: str = ArgProxy(
        '--type',
        required=True,
        choices=['release', 'hotfix'],
        help='verification type',
    )
    m_dir: str = Arg(
        help='m project directory',
        required=True,
        positional=True,
    )


@command(
    help='assert that we are working on the correct branch',
    model=Arguments,
)
def run(arg: Arguments):
    from m.ci.assert_branch import assert_branch

    return run_main(
        lambda: assert_branch(arg.type, arg.m_dir),
        result_handler=lambda _: None,
    )
